using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Events;

namespace Damage
{
    public class HumanDamageBehaviour : MonoBehaviour
    {
        // yo

        public LimbBehaviour limb;
        public PersonBehaviour person;
        List<Organ> organs = new List<Organ>();
        public enum Part {Head, UpperTorso, LowerTorso, Stomach}
        public Part bodyPart;
        public bool hasOrgan = true;
        public bool fracturedBone;
        public float boneHealth = 500f;
        public float initialBoneHealth;
        public float fractureChance = 0.4f;

        public float tissueHealth;
        public float tissueInitialHealth;
        public float tissueDamageMultiplier;
        float shotHeat;
        public bool vitalOrganDestroyed;

        private ContactPoint2D[] contactBuffer = new ContactPoint2D[8];

        
        void Start()
        {
            Organ brain = ScriptableObject.CreateInstance<Organ>();
            brain.name = "Brain";
            brain.health = 1000f;
            brain.initialHealth = 1000f;
            brain.vital = true;
            brain.chance = .7f;
            brain.damageMultiplier = 2f;

            Organ heart = ScriptableObject.CreateInstance<Organ>();
            heart.name = "Heart";
            heart.health = 900;
            heart.initialHealth = 900;
            heart.vital = true;
            heart.chance = .8f;

            Organ lungs = ScriptableObject.CreateInstance<Organ>();
            lungs.name = "Lungs";
            lungs.health = 300f;
            lungs.initialHealth = 300f;
            lungs.chance = .8f;

            limb = gameObject.GetComponent<LimbBehaviour>();
            person = limb.Person;
            switch (gameObject.name)
            {
                case "Head":
                organs.Add(brain);
                break;
                case "UpperBody":
                organs.Add(heart);
                fractureChance = .5f;
                break;
                case "MiddleBody":
                organs.Add(lungs);
                fractureChance = .25f;
                boneHealth = 30f;
                break;
                case "LowerBody":
                fractureChance = .15f;
                boneHealth = 25f;
                break;
                case "UpperLegFront":
                boneHealth = 200f;
                break;
                case "UpperLeg":
                boneHealth = 200f;
                break;
                default:
                hasOrgan = false;
                break;
            }
            switch (limb.RoughClassification)
            {
                case LimbBehaviour.BodyPart.Arms:
                tissueHealth = 250f;
                tissueDamageMultiplier = 0.1f;
                break;
                case LimbBehaviour.BodyPart.Legs:
                tissueHealth = 250f;
                tissueDamageMultiplier = 0.1f;
                break;
                case LimbBehaviour.BodyPart.Head:
                tissueHealth = 500f;
                tissueDamageMultiplier = 0.1f;
                break;
                case LimbBehaviour.BodyPart.Torso:
                tissueHealth = 800f;
                tissueDamageMultiplier = 0.1f;
                break;
            }
            initialBoneHealth = boneHealth;
            tissueInitialHealth = tissueHealth;
        }
        void Update()
        {
            if (GetComponent<LifePoison>())
            {
                HealBone();
                if (hasOrgan)
                {
                    foreach (Organ organ in organs)
                    {
                        RepairOrgan(organ);
                    }
                }
                vitalOrganDestroyed = false;
                tissueHealth = tissueInitialHealth;
            }
            if (hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    if (GetComponent<AcidPoison>())
                        organ.Damage(organ.initialHealth * (Time.deltaTime / 5));
                    if (!organ.destroyed)
                    {
                        if (limb.CirculationBehaviour.BloodFlow < 0.5f)
                        {
                            organ.Damage(Time.deltaTime);
                            //if (organ.destroyed)
                                //ModAPI.Notify("Bleedout");
                        }
                        if (person.OxygenLevel <= 0)
                        {
                            organ.Damage(Time.deltaTime * 2);
                            //if (organ.destroyed)
                                //ModAPI.Notify("Suffocation");
                        }
                        if (limb.PhysicalBehaviour.Temperature >= limb.DiscomfortingHeatTemperature || limb.PhysicalBehaviour.Temperature <= limb.BodyTemperature - 30)
                            organ.Damage(3f/1000f);
                        if (limb.PhysicalBehaviour.Temperature > limb.PhysicalBehaviour.Properties.BurningTemperatureThreshold || limb.PhysicalBehaviour.Temperature < limb.FreezingTemperature)
                            organ.Damage(0.5f);
                    }
                    if (organ.vital)
                    {
                        limb.Health = organ.health;
                        if (!organ.destroyed)
                        {
                            if (limb.IsDismembered || GetComponent<DeathPoison>())
                                KillBody(); 
                            if (limb.IsLethalToBreak && limb.Broken)
                                KillBody();
                        }
                        if (organ.destroyed || vitalOrganDestroyed)
                        {
                            limb.CirculationBehaviour.IsPump = false;
                            if (!vitalOrganDestroyed)
                                KillBody();
                            vitalOrganDestroyed = true;
                        }
                        else
                        {
                            limb.CirculationBehaviour.IsPump = true;
                            vitalOrganDestroyed = false;
                        }
                    }
                    if (organ.health < organ.initialHealth)
                    {
                        if (organ.name == "Lungs")
                        {
                            // cancel out the regeneration of oxygen
                            // just copied straight from dotpeek lol
                            person.OxygenLevel -= (float) ((double) Mathf.Pow(Mathf.Sin(Time.time * 3f), 16f) * 0.5 / 90.0 * ((double) Time.deltaTime * 60.0));
                            // remove oxygen
                            person.OxygenLevel -= Time.deltaTime / Mathf.Clamp(360 * (organ.health / organ.initialHealth), 60, 360);
                        }
                    }  
                }
            }
        }
        void FixedUpdate()
        {
            shotHeat = 1f;
        }
        public void KillBody()
        {
            //ModAPI.Notify("Death");
            foreach (LimbBehaviour limb in limb.Person.Limbs)
            {
                HumanDamageBehaviour damageBehaviour = limb.GetComponent<HumanDamageBehaviour>();
                damageBehaviour.vitalOrganDestroyed = true;
                if (damageBehaviour.hasOrgan)
                {
                    foreach (Organ organ in damageBehaviour.organs)
                    {
                        organ.health = 0;
                        organ.destroyed = true;
                    }
                }
            }
        }
        private void OnFragmentHit(float force)
        {
            float fracturerng = UnityEngine.Random.value;
            if (fracturerng < fractureChance)
                FractureBone(force * 2f);
            if (hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    float rng = UnityEngine.Random.value;
                    DamageOrgan(force * 4f, organ, false, rng);
                }
            }
            //ModAPI.Notify("Fragment hit with force of " + force);
        }
        private void Stabbed(Stabbing stab)
        {
            bool syringe = false;
            List<Component> components = new List<Component>();
            foreach (var component in stab.stabber.GetComponents<Component>())
            {
                if (component.GetType().IsSubclassOf(typeof(SyringeBehaviour)) && !syringe)
                    syringe = true;
            }
            if (syringe == true)
                return;
            float artery = UnityEngine.Random.value;
            if (hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    float damage = UnityEngine.Random.Range(0, organ.initialHealth * 0.5f);
                    float rng = UnityEngine.Random.value * 0.85f;
                    if (organ.name == "Brain")
                        damage = 50f;
                    DamageOrgan(damage, organ, false, rng);
                }
            }
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Arms && artery < 0.3f)
                HitArtery(0.5f);
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Legs && artery < 0.2f)
                HitArtery(1f);
        }
        private void Shot(global::Shot shot)
        {
            shotHeat += 2;
            float damage = shot.damage;
            float fracturerng = UnityEngine.Random.value;
            float fragmentrng = UnityEngine.Random.value;
            if (!fracturedBone && fracturerng < fractureChance)
                FractureBone(damage);
            if (damage > 10 && UnityEngine.Random.value < 0.75f)
            {
                damage *= UnityEngine.Random.Range(1f, 1.5f);
                //ModAPI.Notify("Yawing");
            }
            if (hasOrgan)
            {
                foreach (Organ organ in organs)
                {
                    float organrng = UnityEngine.Random.value;
                    DamageOrgan(damage, organ, false, organrng);
                }
            }
            foreach (LimbBehaviour otherLimb in limb.ConnectedLimbs)
            {
                float rng = UnityEngine.Random.value;
                if (limb.Health <= 0)
                    otherLimb.Damage(shot.damage / 5);
                if (rng < 0.2f)
                {   
                    if (limb.Health > 0)
                        otherLimb.Damage(shot.damage / 20);
                    if (otherLimb.RoughClassification == limb.RoughClassification && otherLimb.gameObject.GetComponent<HumanDamageBehaviour>().hasOrgan)
                    {
                        List<Organ> otherOrgans = otherLimb.gameObject.GetComponent<HumanDamageBehaviour>().organs;
                        foreach (Organ organ in otherOrgans)
                        {
                            float organrng = UnityEngine.Random.value;
                            DamageOrgan(damage / 5, organ, false, organrng);
                        }
                    }
                }
            }
            if (fragmentrng < 0.05f)
            {
                FragmentDamage(UnityEngine.Random.Range(2, 3), Mathf.Clamp(damage / 5, 0, 25));
                //ModAPI.Notify("Bullet fragmentation");
            }
            if (shot.damage >= (tissueHealth / shotHeat))
                DamageTissue(shot.damage, 1);
            float artery = UnityEngine.Random.value;
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Arms && artery < 0.3f)
                HitArtery(0.5f);
            if (limb.RoughClassification == LimbBehaviour.BodyPart.Legs && artery < 0.2f)
                HitArtery(1f);
            StartCoroutine(EnterShock(20f));
        }
        public IEnumerator EnterShock(float time)
        {
            while(time > 0)
            {
                limb.Person.AdrenalineLevel = 20f;
                time -= Time.deltaTime;
                yield return null;
            }
        }
        private void ExitShot(global::Shot shot)
        {
            if (shot.damage >= (tissueHealth / shotHeat))
                DamageTissue(shot.damage, 2);
        }
        public void DamageTissue(float damage, float multiplier)
        {
            damage *= multiplier * shotHeat;
            tissueHealth -= damage * tissueDamageMultiplier;
            //ModAPI.Notify("Tissue damaged by " + damage * tissueDamageMultiplier + " damage. Tissue health now at " + tissueHealth);
            if (tissueHealth <= 0)
            {
                limb.BreakingThreshold = 0.0f;
                //ModAPI.Notify("Dismemberment");
            }
        }
        public void HitArtery(float severity)
        {
            limb.CirculationBehaviour.BloodLossRateMultiplier += severity;
            //ModAPI.Notify("Artery broken");
        }
        public void FractureBone(float damage)
        {
            if (boneHealth <= 0)
                return;
            if (damage < 6)
            {
                limb.Damage(damage);
                return;
            }
            float random = UnityEngine.Random.value;
            // just guessed these numbers
            int fragmentCount = UnityEngine.Random.Range(5, 18);

            if (random < fractureChance || damage < 10)
            {
                float fragmentDamage = Mathf.Clamp(damage, 0, initialBoneHealth / 5);
                limb.BaseStrength /= 2;
                FragmentDamage(fragmentCount, fragmentDamage / fragmentCount);
                boneHealth -= fragmentDamage;
                //ModAPI.Notify("Fractured bone. " + "Health at " + boneHealth);
            }
            else
            {
                limb.BaseStrength = 0;
                float previousBoneHealth = boneHealth;
                boneHealth -= 100;
                boneHealth = Mathf.Clamp(boneHealth, 0, Mathf.Infinity);
                float fragmentDamage = previousBoneHealth - boneHealth;
                float count = fragmentCount * UnityEngine.Random.Range(1, 2);
                FragmentDamage(count, fragmentDamage / count);
                //ModAPI.Notify("Shattered bone. " + "Health at " + boneHealth);
            }
            boneHealth = Mathf.Clamp(boneHealth, 0, Mathf.Infinity);
        }
        public void FragmentDamage(float amount, float damage)
        {
            int neighboringLimbs = limb.ConnectedLimbs.Count;
            amount /= neighboringLimbs;
            foreach (LimbBehaviour otherLimb in limb.ConnectedLimbs)
            {
                otherLimb.Damage(damage * amount);

                HumanDamageBehaviour damageBehaviour;
                bool has;
                CheckForOrgan(out has, out damageBehaviour);
                if (has)
                {
                    foreach (Organ organ in damageBehaviour.organs)
                    {
                        float rng = UnityEngine.Random.value;
                        damageBehaviour.DamageOrgan((damage * amount) * rng, organ, true, 0);
                    }
                }
            }
        }
        public void CheckForOrgan(out bool o, out HumanDamageBehaviour b)
        {
            HumanDamageBehaviour damageBehaviour = limb.gameObject.GetComponent<HumanDamageBehaviour>();
            o = damageBehaviour.hasOrgan;
            b = damageBehaviour;
        }
        public void DamageOrgan(float damage, Organ organ, bool ignoreMultiplier, float rng)
        { 
            if (!ignoreMultiplier)
                damage *= organ.damageMultiplier;
            if (!organ.destroyed && rng < organ.chance)
            {
                organ.Damage(damage);
                //ModAPI.Notify(organ.name + " damaged by " + damage + ". Health is now " + organ.health);
                //if (organ.destroyed)
                    //ModAPI.Notify(organ.name + " destroyed.");
            }
        }
        public void HealBone()
        {
            boneHealth = initialBoneHealth;
            limb.BaseStrength = 5f;
        }
        public void RepairOrgan(Organ organ)
        {
            organ.health = organ.initialHealth;
            organ.destroyed = false;
        }
        void OnCollisionEnter2D(Collision2D collision)
        {
            // ctrl v
            float impulse = ((IEnumerable<ContactPoint2D>)this.contactBuffer).Take<ContactPoint2D>(collision.GetContacts(this.contactBuffer)).Max<ContactPoint2D>((Func<ContactPoint2D, float>)(c => c.normalImpulse));
            if (impulse > .8f)
            {
                float rng = UnityEngine.Random.value;
                float fracturerng = UnityEngine.Random.value;
                if (hasOrgan)
                {
                    foreach (Organ organ in organs)
                    {
                        DamageOrgan(impulse * 10, organ, false, rng * 0.8f);
                    }
                }
                if (fracturerng < .5f)
                {
                    FractureBone(impulse * 2);
                }
            }
        }
    }
    public class Organ : ScriptableObject
    {
        public string name;
        public float health;
        public float initialHealth;
        public float chance;
        public float damageMultiplier = 1f;
        public bool destroyed;
        public bool vital;

        public void Damage(float damage)
        {
            health -= damage;
            if (health <= 0)
                destroyed = true;
        }
    }
}